"""Support sorting for SQLAlchemy ORM."""

import typing as t

from sqlalchemy import Column, sql

from ..sorting import Sort, Sorting


TCOLLECTION = t.TypeVar('TCOLLECTION', bound=sql.Select)


class SASort(Sort):
    """Sorter for Peewee."""

    def apply(self, collection: TCOLLECTION, desc: bool = False, **options) -> TCOLLECTION:
        """Sort the collection."""
        collection.append_order_by(self.field if not desc else self.field.desc())
        return collection


class SASorting(Sorting):
    """Sort Peewee ORM Queries."""

    MUTATE_CLASS = SASort

    def convert(self, obj: t.Union[str, Column, SASort], **meta):
        """Prepare sorters."""
        from . import SARESTHandler

        if isinstance(obj, SASort):
            return obj

        handler = t.cast(SARESTHandler, self.handler)

        if isinstance(obj, Column):
            name, field = obj.name, obj

        else:
            name = obj
            field = meta.get('field', handler.meta.table.c.get(name))

        if field is not None:
            sort = self.MUTATE_CLASS(name, field=field, **meta)
            if sort.meta.get('default'):
                self.default.append(sort)

            return sort

    def sort_default(self, collection: sql.Select) -> sql.Select:
        """Sort collection by default."""
        return collection.order_by(*[
            sort.field.desc() if sort.meta['default'] == 'desc' else sort.field
            for sort in self.default
        ])



"""REST helpers for Muffin Framework."""

__version__ = "3.3.3"
__project__ = "muffin-rest"
__author__ = "Kirill Klenov <horneds@gmail.com>"
__license__ = "MIT"

# Default query params
FILTERS_PARAM = 'where'
LIMIT_PARAM = 'limit'
OFFSET_PARAM = 'offset'
SORT_PARAM = 'sort'


from .api import API                # noqa
from .handler import RESTHandler    # noqa
from .errors import APIError        # noqa

# Just an alias to support legacy style
Api = API


# Support Peewee ORM
try:
    from .peewee import PWRESTHandler
    from .peewee.filters import PWFilter, PWFilters
    from .peewee.sorting import PWSort, PWSorting
except ImportError:
    pass


# Support SQLAlchemy ORM
try:
    from .sqlalchemy import SARESTHandler
    from .sqlalchemy.filters import SAFilter, SAFilters
    from .sqlalchemy.sorting import SASort, SASorting
except ImportError:
    pass


# Support Mongo ORM
try:
    from .mongo import MongoRESTHandler
    from .mongo.filters import MongoFilter, MongoFilters
    from .mongo.sorting import MongoSort, MongoSorting
except ImportError:
    pass

# pylama:ignore=W0611



"""Support sorting for Mongo ORM."""

import typing as t

from ..sorting import Sort, Sorting

from .utils import MongoChain


TCOLLECTION = t.TypeVar('TCOLLECTION', bound=MongoChain)


class MongoSort(Sort):
    """Sorter for Peewee."""

    def apply(self, collection, desc: bool = False, **options):
        """Sort the collection."""
        collection.sorting.append((self.field, -1 if desc else 1))
        return collection


class MongoSorting(Sorting):
    """Sort Peewee ORM Queries."""

    MUTATE_CLASS = MongoSort

    def sort_default(self, collection: TCOLLECTION) -> TCOLLECTION:  # type: ignore
        """Sort collection by default."""
        return collection.sort(
            [(sort.name, -1 if sort.meta['default'] == 'desc' else 1) for sort in self.default])



"""Implement a base class for API."""

import dataclasses as dc
import typing as t
from pathlib import Path

import muffin
from http_router import Router
from muffin.utils import to_awaitable

from .openapi import render_openapi


AUTH = t.TypeVar('AUTH', bound=t.Callable[[muffin.Request], t.Awaitable])
REDOC_TEMPLATE = Path(__file__).parent.joinpath('redoc.html').read_text()
SWAGGER_TEMPLATE = Path(__file__).parent.joinpath('swagger.html').read_text()


@dc.dataclass
class API:
    """Initialize an API."""

    def __init__(self, app: muffin.Application = None, prefix: str = '',
                 openapi: bool = True, *, servers: t.List = None, **openapi_info):
        """Post initialize the API if we have an application already."""
        self.app = app
        self.prefix = prefix

        self.openapi = openapi
        self.openapi_options: t.Dict[str, t.Any] = {'info': openapi_info}
        if servers:
            self.openapi_options['servers'] = servers

        self.authorize: t.Callable[[muffin.Request], t.Awaitable] = to_awaitable(lambda r: True)
        self.router = Router()

        if app:
            self.setup(app, prefix=prefix)

    def __repr__(self):
        """Stringify the API."""
        return f"<API { self.prefix }>"

    def authorization(self, auth: AUTH) -> AUTH:
        """Bind an authorization flow to self API."""
        self.authorize = auth
        return auth

    @property
    def logger(self):
        """Proxy the application's logger."""
        if self.app is None:
            raise RuntimeError('API must be binded to an app.')

        return self.app.logger

    def setup(self, app: muffin.Application, prefix: str = '',
              openapi: bool = None, *, servers: t.List = None, **openapi_info):
        """Initialize the API."""
        self.app = app
        self.prefix = (prefix or self.prefix).rstrip('/')

        # Setup routing
        self.router.trim_last_slash = self.app.router.trim_last_slash
        self.router.validator = self.app.router.validator                   # type: ignore
        self.app.router.route(self.prefix)(self.router)

        if openapi is not None:
            self.openapi = openapi

        if openapi_info:
            self.openapi_options['info'] = openapi_info

        if servers:
            self.openapi_options['servers'] = servers

        # Setup API Docs
        if not self.openapi:
            return

        @self.router.route('/swagger')
        async def swagger(request):
            return SWAGGER_TEMPLATE

        @self.router.route('/redoc')
        async def redoc(request):
            return REDOC_TEMPLATE

        @self.router.route('/openapi.json')
        async def openapi_json(request):
            return render_openapi(self, request=request)

    def route(self, path: t.Union[str, t.Any], *paths: str, **params):
        """Route an endpoint by the API."""
        from .handler import RESTBase

        def wrapper(cb):
            cb._api = self
            return self.router.route(*paths, **params)(cb)

        if isinstance(path, str):
            paths = (path, *paths)
            return wrapper

        # Generate URL paths automatically
        if issubclass(path, RESTBase):
            path._api = self
            return self.router.route(path, *paths, **params)

        raise Exception("Invalid endpoint")  # TODO



"""Support filters for SQLAlchemy ORM."""

import typing as t
from sqlalchemy import Column, sql
import marshmallow as ma

from ..filters import Filter, Filters


TCOLLECTION = t.TypeVar('TCOLLECTION', bound=sql.Select)


class SAFilter(Filter):
    """Custom filter for sqlalchemy."""

    operators = Filter.operators
    operators['$between'] = lambda c, v: c.between(*v)
    operators['$ends'] = lambda c, v: c.endswith(v)
    operators['$ilike'] = lambda c, v: c.ilike(v)
    operators['$in'] = lambda c, v: c.in_(v)
    operators['$like'] = lambda c, v: c.like(v)
    operators['$match'] = lambda c, v: c.match(v)
    operators['$nin'] = lambda c, v: ~c.in_(v)
    operators['$notilike'] = lambda c, v: c.notilike(v)
    operators['$notlike'] = lambda c, v: c.notlike(v)
    operators['$starts'] = lambda c, v: c.startswith(v)

    list_ops = Filter.list_ops + ['$between']

    def __init__(self, name: str, *, field: Column = None,
                 schema_field: ma.fields.Field = None, operator: str = None, **meta):
        """Support custom model fields."""
        self.name = name
        self.field = field
        self.schema_field = schema_field or self.schema_field_cls(
            attribute=field is not None and (field.key or field.name) or name)

        if operator:
            self.default_operator = operator

    def apply(self, collection: sql.Select,
              *ops: t.Tuple[t.Callable, t.Any], **kwargs) -> sql.Select:
        """Apply the filters to SQLAlchemy Select."""
        column = self.field
        if ops and column is not None:
            return self.query(collection, column, *ops, **kwargs)

        return collection

    def query(self, select: sql.Select, column: Column, *ops, **kwargs) -> sql.Select:
        """Filter a select."""
        return select.where(*[op(column, val) for op, val in ops])


class SAFilters(Filters):
    """Bind SAfilter class."""

    MUTATE_CLASS = SAFilter

    def convert(self, obj: t.Union[str, Column, SAFilter], **meta):
        """Convert params to filters."""
        from . import SARESTHandler

        handler = t.cast(SARESTHandler, self.handler)

        if isinstance(obj, SAFilter):
            if obj.field is None:
                obj.field = handler.meta.table.c.get(obj.name)
            return obj

        if isinstance(obj, Column):
            name = obj.name
            field = obj

        else:
            name = obj
            field = meta.pop('field', None) or name
            if isinstance(field, str):
                field = handler.meta.table.c.get(field)

        schema_field = meta.pop('schema_field', None)
        if schema_field is None and field is not None:
            schema_field = handler.meta.Schema._declared_fields.get(field.name)

        return self.MUTATE_CLASS(name, field=field, schema_field=schema_field, **meta)



"""Support filters for Peewee ORM."""

import typing as t
from peewee import Query, Field
import marshmallow as ma

from ..filters import Filter, Filters


class PWFilter(Filter):
    """Support Peewee."""

    operators = Filter.operators
    operators['$in'] = lambda f, v: f << v
    operators['$none'] = lambda f, v: f >> v
    operators['$like'] = lambda f, v: f % v
    operators['$ilike'] = lambda f, v: f ** v
    operators['$contains'] = lambda f, v: f.contains(v)
    operators['$starts'] = lambda f, v: f.startswith(v)
    operators['$ends'] = lambda f, v: f.endswith(v)
    operators['$between'] = lambda f, v: f.between(*v)
    operators['$regexp'] = lambda f, v: f.regexp(v)

    list_ops = Filter.list_ops + ['$between']

    def __init__(self, name: str, *, field: Field = None,
                 schema_field: ma.fields.Field = None, operator: str = None, **meta):
        """Support custom model fields."""
        self.name = name
        self.field = field
        self.schema_field = schema_field or self.schema_field_cls(
            attribute=field and field.name or name)
        if operator:
            self.default_operator = operator

    def apply(self, collection: Query, *ops: t.Tuple[t.Callable, t.Any], **kwargs) -> Query:
        """Apply the filters to Peewee QuerySet.."""
        if ops:
            return self.query(collection, self.field, *ops, **kwargs)
        return collection

    def query(self, query: Query, column: Field, *ops: t.Tuple, **kwargs) -> Query:
        """Filter a query."""
        if isinstance(column, Field):
            return query.where(*[op(column, val) for op, val in ops])

        return query


class PWFilters(Filters):
    """Bind Peewee filter class."""

    MUTATE_CLASS = PWFilter

    def convert(self, obj: t.Union[str, Field, PWFilter], **meta):
        """Convert params to filters."""
        from . import PWRESTHandler

        handler = t.cast(PWRESTHandler, self.handler)
        if isinstance(obj, PWFilter):
            if obj.field is None:
                obj.field = handler.meta.model._meta.fields.get(obj.name)
            return obj

        if isinstance(obj, Field):
            name = obj.name
            field = obj

        else:
            name = obj
            field = meta.pop('field', None) or name
            if isinstance(field, str):
                field = handler.meta.model._meta.fields.get(field)

        schema_field = meta.pop('schema_field', None)
        if schema_field is None and field:
            schema_field = handler.meta.Schema._declared_fields.get(field.name)
        return self.MUTATE_CLASS(name, field=field, schema_field=schema_field, **meta)



"""Mongo DB support."""

import typing as t

import bson
import marshmallow as ma
import muffin
from motor import motor_asyncio as motor

from ..handler import RESTHandler, RESTOptions
from ..errors import APIError

from .schema import MongoSchema
from .utils import MongoChain
from .filters import MongoFilters
from .sorting import MongoSorting


class MongoRESTOptions(RESTOptions):
    """Support Mongo DB."""

    filters_cls: t.Type[MongoFilters] = MongoFilters
    sorting_cls: t.Type[MongoSorting] = MongoSorting
    schema_base: t.Type[MongoSchema] = MongoSchema

    aggregate: t.Optional[t.List] = None  # Support aggregation. Set to pipeline.
    collection_id: str = '_id'
    collection: t.Optional[motor.AsyncIOMotorCollection] = None

    base_property: str = 'collection'

    if t.TYPE_CHECKING:
        Schema: t.Type[MongoSchema]

    def setup(self, cls):
        """Prepare meta options."""
        self.name = self.name or self.collection.name.lower()
        super(MongoRESTOptions, self).setup(cls)


class MongoRESTHandler(RESTHandler):
    """Support Mongo DB."""

    meta: MongoRESTOptions
    meta_class: t.Type[MongoRESTOptions] = MongoRESTOptions

    async def prepare_collection(self, request: muffin.Request) -> MongoChain:
        """Initialize Peeewee QuerySet for a binded to the resource model."""
        return MongoChain(self.meta.collection)

    async def paginate(self, request: muffin.Request, *, limit: int = 0,
                       offset: int = 0) -> t.Tuple[motor.AsyncIOMotorCursor, int]:
        """Paginate collection."""
        if self.meta.aggregate:
            pipeline_all = self.meta.aggregate + [{'$skip': offset}, {'$limit': limit}]
            pipeline_num = self.meta.aggregate + [{'$group': {
                self.meta.collection_id: None, 'total': {'$sum': 1}}}]
            counts = list(self.collection.aggregate(pipeline_num))
            return (
                self.collection.aggregate(pipeline_all),
                counts and counts[0]['total'] or 0
            )
        total = await self.collection.count()
        return self.collection.skip(offset).limit(limit), total

    async def get(self, request, *, resource=None):
        """Get resource or collection of resources."""
        if resource is not None and resource != '':
            return await self.dump(request, resource, many=False)

        docs = await self.collection.to_list(None)
        return await self.dump(request, docs, many=True)

    async def prepare_resource(self, request: muffin.Request) -> t.Optional[dict]:
        """Load a resource."""
        pk = request['path_params'].get(self.meta.name_id)
        if not pk:
            return None

        try:
            return await self.collection.find_one({self.meta.collection_id: bson.ObjectId(pk)})
        except bson.errors.InvalidId:
            raise APIError.NOT_FOUND()

    async def get_schema(self, request: muffin.Request, resource=None) -> ma.Schema:
        """Initialize marshmallow schema for serialization/deserialization."""
        return self.meta.Schema(
            instance=resource,
            only=request.url.query.get('schema_only'),
            exclude=request.url.query.get('schema_exclude', ()),
        )

    async def save(self, request: muffin.Request,  # type: ignore
                   resource: t.Union[dict, t.List[dict]]):
        """Save the given resource.

        Supports batch saving.
        """
        docs = resource if isinstance(resource, t.Sequence) else [resource]
        for doc in docs:
            if doc.get(self.meta.collection_id):
                await self.collection.replace_one(
                    {self.meta.collection_id: doc[self.meta.collection_id]}, doc)

            else:
                res = await self.meta.collection.insert_one(doc)  # type: ignore
                doc[self.meta.collection_id] = res.inserted_id

        return resource

    async def remove(self, request: muffin.Request, *, resource: dict = None):
        """Remove the given resource(s)."""
        oids = [resource[self.meta.collection_id]] if resource else await request.data()
        if not oids:
            raise APIError.NOT_FOUND()

        if not isinstance(oids, list):
            raise APIError.BAD_REQUEST()

        oids = [bson.ObjectId(_id) for _id in oids]
        await self.meta.collection.delete_many(  # type: ignore
            {self.meta.collection_id: {'$in': oids}})

    delete = remove  # noqa



"""Implement sorting."""

import typing as t

from muffin import Request
from muffin.handler import Handler

from . import SORT_PARAM
from .utils import Mutate, Mutator, TCOLLECTION


class Sort(Mutate):
    """Sort a collection."""

    def apply(self, collection, desc: bool = False, **options):
        """Sort the collection."""
        return sorted(collection, key=lambda obj: getattr(obj, self.name), reverse=desc)


class Sorting(Mutator):

    """Build sorters for handlers."""

    MUTATE_CLASS = Sort

    def __init__(self, handler: Handler, params: t.Sequence):
        """Initialize the sorting."""
        self.default: t.List = []
        super(Sorting, self).__init__(handler, params)

    def convert(self, obj, **meta):
        """Prepare sorters."""
        sort = super(Sorting, self).convert(obj, **meta)
        if sort.meta.get('default'):
            self.default.append(sort)
        return sort

    def apply(self, request: Request, collection: TCOLLECTION, **options) -> TCOLLECTION:
        """Sort the given collection."""
        data = request.url.query.get(SORT_PARAM)
        if data:
            for name, desc in to_sort(data.split(',')):
                if name in self.mutations:
                    collection = self.mutations[name].apply(collection, desc)

        elif self.default:
            return self.sort_default(collection)

        return collection

    def sort_default(self, collection: TCOLLECTION) -> TCOLLECTION:
        """Sort by default."""
        return sorted(collection)  # type: ignore

    @property
    def openapi(self):
        """Prepare OpenAPI params."""
        sorting = list(self.mutations)
        return {
            'name': SORT_PARAM, 'in': 'query', 'style': 'form', 'explode': False,
            'schema': {'type': 'array', 'items': {'type': 'string', 'enum': sorting}},
            'description': ",".join(sorting),
        }


def to_sort(sort_params: t.Sequence[str]) -> t.Generator[t.Tuple[str, bool], None, None]:
    """Generate sort params."""
    for name in sort_params:
        n, desc = name.strip('-'), name.startswith('-')
        if n:
            yield n, desc



"""Base class for API REST Handlers."""
from __future__ import annotations

import abc
import inspect
import typing as t

import marshmallow as ma
from asgi_tools.response import parse_response
from muffin import Request
from muffin.typing import JSONType
from muffin.handler import Handler, HandlerMeta

from . import LIMIT_PARAM, OFFSET_PARAM, openapi
from .api import API
from .errors import APIError
from .filters import Filters, Filter
from .sorting import Sorting, Sort


T = t.TypeVar('T')


class RESTOptions:
    """Handler Options."""

    name: t.Optional[str] = None
    name_id: str = 'id'

    # limit: Paginate results (set to None for disable pagination)
    limit: int = 0
    limit_max: int = 0

    # Base class for filters
    filters: Filters
    filters_cls: t.Type[Filters] = Filters

    # Base class for sorting
    sorting: Sorting
    sorting_cls: t.Type[Sorting] = Sorting

    # Auto generation for schemas
    Schema: t.Type[ma.Schema]
    schema_base: t.Type[ma.Schema] = ma.Schema
    schema_fields: t.Dict = {}
    schema_meta: t.Dict = {}
    schema_unknown: str = ma.EXCLUDE

    base_property: str = 'name'

    def __init__(self, cls):
        """Inherit meta options."""
        for base in reversed(cls.mro()):
            if hasattr(base, "Meta"):
                for k, v in base.Meta.__dict__.items():
                    if not k.startswith('_'):
                        setattr(self, k, v)

        if getattr(self, self.base_property, None) is not None:
            self.setup(cls)

    def setup(self, cls):
        """Setup the options."""
        if not self.Schema:
            self.Schema = type(
                self.name.title() + 'Schema', (self.schema_base,),
                dict(self.schema_fields, Meta=self.setup_schema_meta(cls)))

        if not self.limit_max:
            self.limit_max = self.limit

    def setup_schema_meta(self, cls):
        """Generate meta for schemas."""
        return type('Meta', (object,), dict({'unknown': self.schema_unknown}, **self.schema_meta))

    def __repr__(self):
        """Represent self as a string."""
        return f"<Options {self.name}>"


class RESTHandlerMeta(HandlerMeta):
    """Create class options."""

    def __new__(mcs, name, bases, params):
        """Prepare options for the handler."""
        cls = super().__new__(mcs, name, bases, params)
        cls.meta = cls.meta_class(cls)

        if getattr(cls.meta, cls.meta_class.base_property, None) is not None:
            cls.meta.filters = cls.meta.filters_cls(cls, cls.meta.filters)
            cls.meta.sorting = cls.meta.sorting_cls(cls, cls.meta.sorting)

        return cls


class RESTBase(Handler, metaclass=RESTHandlerMeta):

    """Load/save resources."""

    if t.TYPE_CHECKING:
        auth: t.Any
        collection: t.Any
        resource: t.Any

    meta: RESTOptions
    meta_class: t.Type[RESTOptions] = RESTOptions
    _api: t.Optional[API] = None

    class Meta:
        """Tune the handler."""

        # Resource filters
        filters: t.Sequence[t.Union[str, t.Tuple[str, str], Filter]] = ()

        # Define allowed resource sorting params
        sorting: t.Sequence[t.Union[str, t.Tuple[str, t.Dict], Sort]] = ()

        # Serialize/Deserialize Schema class
        Schema: t.Optional[t.Type[ma.Schema]] = None

    @classmethod
    def __route__(cls, router, *paths, **params):
        """Bind the class to the given router."""
        methods = params.pop('methods') or cls.methods
        if paths:
            router.bind(cls, *paths, methods=methods, **params)

        else:
            router.bind(cls, f"/{ cls.meta.name }", methods=methods, **params)
            router.bind(cls, f"/{ cls.meta.name }/{{{ cls.meta.name_id }}}",
                        methods=methods, **params)

        for _, method in inspect.getmembers(cls, lambda m: hasattr(m, '__route__')):
            cpaths, cparams = method.__route__
            router.bind(cls, *cpaths, __meth__=method.__name__, **cparams)

        return cls

    async def __call__(self, request: Request, *args, __meth__: str = None, **options) -> t.Any:
        """Dispatch the given request by HTTP method."""
        method = getattr(self, __meth__ or request.method.lower())
        self.auth = await self.authorize(request)
        self.collection = await self.prepare_collection(request)
        resource = await self.prepare_resource(request)
        if resource or request.method != 'GET':
            return await method(request, resource=resource)

        meta = self.meta

        # Filter collection
        if meta.filters:
            self.collection = meta.filters.apply(request, self.collection, **options)

        # Sort collection
        if meta.sorting:
            self.collection = meta.sorting.apply(request, self.collection, **options)

        # Paginate the collection
        headers = {}
        if meta.limit:
            limit, offset = self.paginate_prepare_params(request)
            if limit and offset >= 0:
                self.collection, total = await self.paginate(
                    request, limit=limit, offset=offset)
                headers = self.paginate_prepare_headers(limit, offset, total)

        response = await method(request, resource=resource)
        if headers:
            response = parse_response(response)
            response.headers.update(headers)

        return response

    @property
    def api(self) -> API:
        """Check if the handler is binded to an API."""
        if self._api is None:
            raise Exception('The handler is not routed by any API')  # TODO
        return self._api

    async def authorize(self, request: Request):
        """Default authorization method. Proxy auth to self.api."""
        auth = await self.api.authorize(request)
        if not auth:
            raise APIError.UNAUTHORIZED()
        return auth

    # Prepare data
    # ------------
    @abc.abstractmethod
    async def prepare_collection(self, request: Request) -> t.Any:
        """Prepare a collection of resources. Create queryset, db cursor and etc."""
        raise NotImplementedError

    async def prepare_resource(self, request: Request) -> t.Any:
        """Load a resource."""
        return request['path_params'].get(self.meta.name_id)

    # Paginate
    # --------
    def paginate_prepare_headers(self, limit, offset, total):
        """Prepare pagination headers."""
        return {'x-total': total, 'x-limit': limit, 'x-offset': offset}

    def paginate_prepare_params(self, request: Request) -> t.Tuple[int, int]:
        """Prepare pagination params."""
        meta = self.meta
        query = request.url.query
        limit = query.get(LIMIT_PARAM) or meta.limit
        try:
            return min(abs(int(limit)), meta.limit_max), int(query.get(OFFSET_PARAM, 0))
        except ValueError:
            raise APIError.BAD_REQUEST('Pagination params are invalid')

    @abc.abstractmethod
    async def paginate(self, request: Request, *, limit: int = 0,
                       offset: int = 0) -> t.Tuple[t.Any, int]:
        """Paginate the results."""
        raise NotImplementedError

    # Manage data
    # -----------
    @abc.abstractmethod
    async def save(self, request: Request, *, resource: T = None) -> T:
        """Save the given resource."""
        raise NotImplementedError

    @abc.abstractmethod
    async def remove(self, request: Request, *, resource: t.Any = None):
        """Remove the given resource."""
        raise NotImplementedError

    # Parse data
    # -----------
    async def get_schema(self, request: Request, resource=None) -> ma.Schema:
        """Initialize marshmallow schema for serialization/deserialization."""
        assert self.meta.Schema, 'RESTHandler.meta.Schema is required.'
        query = request.url.query
        return self.meta.Schema(
            only=query.get('schema_only'),
            exclude=query.get('schema_exclude', ()),
        )

    async def parse(self, request: Request):
        """Parse data from the given request."""
        try:
            return await request.data(raise_errors=True)
        except (ValueError, TypeError) as exc:
            raise APIError.BAD_REQUEST(str(exc))

    async def load(self, request: Request, *, resource: t.Any = None) -> t.Any:
        """Load data from request and create/update a resource."""
        data = await self.parse(request)
        schema = await self.get_schema(request, resource=resource)
        if not schema:
            return data

        try:
            resource = schema.load(data, partial=resource is not None, many=isinstance(data, list))  # type: ignore # noqa
        except ma.ValidationError as exc:
            raise APIError.BAD_REQUEST('Invalid data', errors=exc.messages)

        return resource

    async def dump(self, request: Request, response: t.Any, *, many=...) -> JSONType:
        """Serialize the given response."""
        schema = await self.get_schema(request)
        if many is ...:
            many = isinstance(response, t.Sequence)

        return schema.dump(response, many=many) if schema else response

    async def get(self, request, *, resource=None) -> JSONType:
        """Get a resource or a collection of resources.

        Specify a path param to load a resource.
        """
        if resource is not None and resource != '':
            return await self.dump(request, resource)

        return await self.dump(request, self.collection, many=True)

    async def post(self, request, *, resource=None) -> JSONType:
        """Create a resource.

        The method accepts a single resource's data or a list of resources to create.
        """
        resource = await self.load(request, resource=resource)
        resource = await self.save(request, resource=resource)
        return await self.dump(request, resource, many=isinstance(resource, list))

    async def put(self, request, *, resource=None) -> JSONType:
        """Update a resource."""
        if resource is None:
            raise APIError.NOT_FOUND()

        return await self.post(request, resource=resource)

    async def delete(self, request, *, resource=None):
        """Delete a resource."""
        if resource is None:
            raise APIError.NOT_FOUND()

        return await self.remove(request, resource=resource)


class RESTHandler(RESTBase, openapi.OpenAPIMixin):
    """Basic Handler Class."""

    pass


def to_sort(sort_params: t.Sequence[str]) -> t.Generator[t.Tuple[str, bool], None, None]:
    """Generate sort params."""
    for name in sort_params:
        n, desc = name.strip('-'), name.startswith('-')
        if n:
            yield n, desc



"""Support API filters."""
import operator
import typing as t

import marshmallow as ma
from muffin import Request
from asgi_tools._compat import json_loads

from . import FILTERS_PARAM, API
from .utils import Mutate, Mutator, TCOLLECTION


class Filter(Mutate):

    """Base filter class."""

    operators = {
        '<':   operator.lt,
        '$lt': operator.lt,
        '<=':  operator.le,
        '$le': operator.le,
        '>':   operator.gt,
        '$gt': operator.gt,
        '>=':  operator.ge,
        '$ge': operator.ge,
        '==':  operator.eq,
        '$eq': operator.eq,
        '!':   operator.ne,
        '$ne': operator.ne,
        '$in': lambda v, c: v in c,
        '$nin': lambda v, c: v not in c,
    }
    operators['<<'] = operators['$in']

    list_ops = ['$in', '<<']

    schema_field_cls: t.Type[ma.fields.Field] = ma.fields.Raw
    default_operator: str = '$eq'

    def __init__(self, name: str, *, field: t.Any = None,
                 schema_field: ma.fields.Field = None, operator: str = None, **meta):
        """Initialize filter.

        :param name: The filter's name
        :param attr: Column/Property name.
        :param fields: Marshmallow Field instance

        """
        super(Filter, self).__init__(name, **meta)
        self.field = field
        self.schema_field = schema_field or self.schema_field_cls()
        if operator:
            self.default_operator = operator

    def filter(self, collection: t.Any, data: t.Mapping, **kwargs):
        """Filter given collection."""
        try:
            ops = self.parse(data)
        except ma.ValidationError:
            return None, collection

        collection = self.apply(collection, *ops, **kwargs)
        return ops, collection

    def parse(self, data: t.Mapping) -> t.Tuple[t.Tuple[t.Callable, t.Any], ...]:
        """Parse operator and value from filter's data."""
        val = data.get(self.name, ma.missing)
        if not isinstance(val, dict):
            return (self.operators[self.default_operator], self.schema_field.deserialize(val)),

        return tuple(
            (
                self.operators[op],
                (self.schema_field.deserialize(val)) if op not in self.list_ops else [
                    self.schema_field.deserialize(v) for v in val])
            for (op, val) in val.items() if op in self.operators
        )

    def apply(self, collection, *ops: t.Tuple[t.Callable, t.Any], **options):
        """Apply the filter to collection."""
        validator = lambda obj: all(op(get_value(obj, self.name), val) for (op, val) in ops)  # noqa
        return [o for o in collection if validator(o)]


class Filters(Mutator):

    """Build filters for handlers."""

    MUTATE_CLASS = Filter

    def convert(self, obj, **meta):
        """Convert params to filters."""
        if isinstance(obj, self.MUTATE_CLASS):
            return obj

        field = meta.pop('field', None) or obj
        schema_field = meta.pop('schema_field', None)
        if schema_field is None and field:
            schema_field = self.handler.meta.Schema._declared_fields.get(field)
        return self.MUTATE_CLASS(obj, field=field, schema_field=schema_field, **meta)

    def apply(self, request: Request, collection: TCOLLECTION, **options) -> TCOLLECTION:
        """Filter the given collection."""
        data = request.url.query.get(FILTERS_PARAM)
        if data is not None:
            try:
                data = json_loads(data)
                assert isinstance(data, dict)
                mutations = self.mutations
                for name in data:
                    if name in mutations:
                        _, collection = mutations[name].filter(collection, data, **options)

            except (ValueError, TypeError, AssertionError):
                api = t.cast(API, self.handler._api)
                api.logger.warning(f'Invalid filters data: { request.url }')

        return collection

    @property
    def openapi(self):
        """Prepare OpenAPI params."""
        return {
            'name': FILTERS_PARAM, 'in': 'query', 'description': str(self),
            'content': {'application/json': {'schema': {'type': 'object'}}}
        }


def get_value(obj, name):
    """Get value from object by name."""
    if isinstance(obj, dict):
        return obj.get(name)

    return getattr(obj, name, obj)



"""SQLAlchemy Core support."""

from __future__ import annotations  # py37

import typing as t

import marshmallow as ma
import muffin
import sqlalchemy as sa
from marshmallow_sqlalchemy import SQLAlchemyAutoSchema as BaseSQLAlchemyAutoSchema, ModelConverter
from muffin.typing import JSONType
from muffin_databases import Plugin as DB

from ..handler import RESTHandler, RESTOptions
from ..errors import APIError
from .filters import SAFilters
from .sorting import SASorting


# XXX: Monkey patch ModelConverter
ModelConverter._get_field_name = lambda s, prop_or_column: str(prop_or_column.key)


class SQLAlchemyAutoSchema(BaseSQLAlchemyAutoSchema):
    """Allow partial updates for tables."""

    @ma.pre_load
    def fill_defaults(self, data, partial=False, **kwargs):
        """Insert default params for SQLAlchemy because databases don't.

        https://github.com/encode/databases/issues/72
        """
        cols_to_fields = {
            f.attribute or f.name: f for f in self.declared_fields.values()}
        if not partial:
            for column in self.opts.table.columns:
                field = cols_to_fields.get(column.name)
                if not field:
                    continue

                data_key = field.data_key or field.name
                if data_key not in data and column.default is not None:
                    value = column.default.arg
                    if callable(value):
                        value = value(column)

                    data[data_key] = field._serialize(value, field.attribute or field.name, None)

        return data

    @ma.post_load
    def make_instance(self, data, **kwargs):
        """Update a table instance."""
        if self.instance is not None:
            instance = dict(self.instance)
            instance.update(data)
            return instance
        return super(SQLAlchemyAutoSchema, self).make_instance(data, **kwargs)


class SARESTOptions(RESTOptions):
    """Support SQLAlchemy Core."""

    filters_cls: t.Type[SAFilters] = SAFilters
    sorting_cls: t.Type[SASorting] = SASorting

    # Schema auto generation params
    Schema: t.Type[SQLAlchemyAutoSchema]
    schema_base: t.Type[SQLAlchemyAutoSchema] = SQLAlchemyAutoSchema

    table: sa.Table
    table_pk: t.Optional[sa.Column] = None
    database: DB

    base_property = 'table'

    def setup(self, cls):
        """Prepare meta options."""
        if self.database is None:
            raise ValueError("'SARESTHandler.Meta.database' is required")

        self.name = self.name or self.table.name
        self.table_pk = self.table_pk or self.table.c.id

        super(SARESTOptions, self).setup(cls)

    def setup_schema_meta(self, cls):
        """Prepare a schema."""
        return type('Meta', (object,), dict({
            'unknown': self.schema_unknown, 'table': self.table,
            'include_fk': True, 'dump_only': (self.name_id,)
        }, **self.schema_meta))


class SARESTHandler(RESTHandler):
    """Support SQLAlchemy Core."""

    meta: SARESTOptions
    meta_class: t.Type[SARESTOptions] = SARESTOptions

    async def prepare_collection(self, request: muffin.Request) -> sa.sql.Select:
        """Initialize Peeewee QuerySet for a binded to the resource model."""
        return self.meta.table.select()

    async def paginate(self, request: muffin.Request, *, limit: int = 0,
                       offset: int = 0) -> t.Tuple[sa.sql.Select, int]:
        """Paginate the collection."""
        qs = sa.select([sa.func.count()]).select_from(self.collection.order_by(None))
        total = await self.meta.database.fetch_val(qs)
        return self.collection.offset(offset).limit(limit), total

    async def get(self, request, *, resource=None) -> JSONType:
        """Get resource or collection of resources."""
        if resource is not None and resource != '':
            return await self.dump(request, resource, many=False)

        rows = await self.meta.database.fetch_all(self.collection)
        return await self.dump(request, rows, many=True)

    async def prepare_resource(self, request: muffin.Request) -> t.Optional[dict]:
        """Load a resource."""
        pk = request['path_params'].get(self.meta.name_id)
        if not pk:
            return None

        qs = self.collection.where(self.meta.table_pk == pk)
        resource = await self.meta.database.fetch_one(qs)
        if resource is None:
            raise APIError.NOT_FOUND('Resource not found')
        return dict(resource)

    async def get_schema(self, request: muffin.Request, resource=None) -> ma.Schema:
        """Initialize marshmallow schema for serialization/deserialization."""
        return self.meta.Schema(
            instance=resource,
            only=request.url.query.get('schema_only'),
            exclude=request.url.query.get('schema_exclude', ()),
        )

    async def save(self, request: muffin.Request,  # type: ignore
                   resource: t.Union[dict, t.List[dict]]):
        """Save the given resource.

        Supports batch saving.
        """
        resources = resource if isinstance(resource, t.Sequence) else [resource]
        table_pk = t.cast(sa.Column, self.meta.table_pk)
        for res in resources:
            if res.get(table_pk.name):
                update = self.meta.table.update().where(
                    table_pk == res[table_pk.name])
                await self.meta.database.execute(update, res)

            else:
                insert = self.meta.table.insert()
                res[table_pk.name] = await self.meta.database.execute(insert, res)

        return resource

    async def remove(self, request: muffin.Request, *, resource: dict = None):
        """Remove the given resource."""
        table_pk = t.cast(sa.Column, self.meta.table_pk)
        pks = [resource[table_pk.name]] if resource else await request.data()
        if not pks:
            raise APIError.NOT_FOUND()

        delete = self.meta.table.delete(table_pk.in_(pks))
        await self.meta.database.execute(delete)

    delete = remove  # noqa



"""Support filters for Mongo."""

import typing as t

from ..filters import Filter, Filters


class MongoFilter(Filter):
    """Custom filter for sqlalchemy."""

    operators = {
        '$eq': lambda n, v: ('$eq', v),
        '$ge': lambda n, v: ('$ge', v),
        '$gt': lambda n, v: ('$gt', v),
        '$in': lambda n, v: ('$in', v),
        '$le': lambda n, v: ('$le', v),
        '$lt': lambda n, v: ('$lt', v),
        '$ne': lambda n, v: ('$ne', v),
        '$nin': lambda n, v: ('$nin', v),
        '$starts': lambda n, v: ('$regex', f"^{ v }"),
        '$ends': lambda n, v: ('$regex', f"{ v }$"),
    }

    def apply(self, collection, *ops: t.Tuple[t.Callable, t.Any], **options):
        """Apply the filter."""
        return collection.find({self.field: dict(op(self.name, v) for op, v in ops)})


class MongoFilters(Filters):
    """Bind MongoFilter class."""

    MUTATE_CLASS = MongoFilter



"""Support for Peewee ORM (https://github.com/coleifer/peewee)."""

from __future__ import annotations

import typing as t

import marshmallow as ma
import muffin
import peewee as pw
from apispec.ext.marshmallow import MarshmallowPlugin
from marshmallow_peewee import ModelSchema, ForeignKey

from ..handler import RESTBase, RESTOptions
from ..errors import APIError

from .openapi import PeeweeOpenAPIMixin
from .filters import PWFilters
from .sorting import PWSorting


# XXX: Patch apispec.MarshmallowPlugin to support ForeignKeyField
MarshmallowPlugin.Converter.field_mapping[ForeignKey] = ("integer", None)


class PWRESTOptions(RESTOptions):
    """Support Peewee."""

    model: pw.Model
    model_pk: t.Optional[pw.Field] = None

    # Base filters class
    filters_cls: t.Type[PWFilters] = PWFilters

    # Base sorting class
    sorting_cls: t.Type[PWSorting] = PWSorting

    Schema: t.Type[ModelSchema]

    # Schema auto generation params
    schema_base: t.Type[ModelSchema] = ModelSchema

    base_property: str = 'model'

    def setup(self, cls):
        """Prepare meta options."""
        self.name = self.name or self.model._meta.table_name.lower()
        self.model_pk = self.model_pk or self.model._meta.primary_key
        super(PWRESTOptions, self).setup(cls)

    def setup_schema_meta(self, cls):
        """Prepare a schema."""
        return type('Meta', (object,), dict(
            {'unknown': self.schema_unknown, 'model': self.model}, **self.schema_meta))


class PWRESTBase(RESTBase):
    """Support Peeweee."""

    collection: pw.Query
    resource: pw.Model
    meta: PWRESTOptions
    meta_class: t.Type[PWRESTOptions] = PWRESTOptions

    async def prepare_collection(self, request: muffin.Request) -> pw.Query:
        """Initialize Peeewee QuerySet for a binded to the resource model."""
        return self.meta.model.select()

    async def prepare_resource(self, request: muffin.Request) -> t.Optional[pw.Model]:
        """Load a resource."""
        pk = request['path_params'].get(self.meta.name_id)
        if not pk:
            return None

        try:
            return self.collection.where(self.meta.model_pk == pk).get()
        except self.meta.model.DoesNotExist:
            raise APIError.NOT_FOUND('Resource not found')

    async def paginate(self, request: muffin.Request, *, limit: int = 0,
                       offset: int = 0) -> t.Tuple[pw.Query, int]:
        """Paginate the collection."""
        cqs = self.collection.order_by()
        if cqs._group_by:
            cqs._select = cqs._group_by
        return self.collection.offset(offset).limit(limit), cqs.count()

    async def save(self, request: muffin.Request,  # type: ignore
                   resource: t.Union[pw.Model, t.List[pw.Model]]):
        """Save the given resource.

        Supports batch saving.
        """
        for obj in (resource if isinstance(resource, list) else [resource]):
            obj.save()
        return resource

    async def remove(self, request: muffin.Request, *, resource: pw.Model = None):
        """Remove the given resource."""
        if resource:
            resources = [resource]

        else:
            data = await request.data()
            if not data:
                return

            model_pk = t.cast(pw.Field, self.meta.model_pk)
            resources = list(self.collection.where(model_pk << data))

        if not resources:
            raise APIError.NOT_FOUND()

        for resource in resources:
            resource.delete_instance()

    delete = remove  # noqa

    async def get_schema(self, request: muffin.Request, resource=None) -> ma.Schema:
        """Initialize marshmallow schema for serialization/deserialization."""
        return self.meta.Schema(
            instance=resource,
            only=request.url.query.get('schema_only'),
            exclude=request.url.query.get('schema_exclude', ()),
        ) if self.meta.Schema else None


class PWRESTHandler(PWRESTBase, PeeweeOpenAPIMixin):  # type: ignore
    """Support peewee."""

    pass



"""Helpers to raise API errors as JSON responses."""
import json
import typing as t
from http import HTTPStatus

from muffin import ResponseError
from muffin.typing import JSONType


class APIError(ResponseError):
    """JSON Response."""

    def __init__(self, content: t.Union[str, JSONType] = None, *,
                 status_code: int = HTTPStatus.BAD_REQUEST.value, **json_data):
        """Create JSON with errors information."""
        response = {'error': True, 'message': HTTPStatus(status_code).description}

        if isinstance(content, t.Mapping):
            response = content

        elif content is not None:
            response['message'] = str(content)

        if json_data:
            response.update(json_data)

        return super(APIError, self).__init__(json.dumps(response), status_code=status_code,
                                              headers={'content-type': 'application/json'})
